import argparse
import pathlib
import sys
from lark import __version__ as lark_version
from gersemi.runner import run, error, SUCCESS, FAIL
from gersemi.__version__ import __title__, __version__


def create_argparser():
    parser = argparse.ArgumentParser(
        description="A formatter to make your CMake code the real treasure.",
        prog="gersemi",
    )
    parser.add_argument(
        "-c",
        "--check",
        dest="check_formatting",
        default=False,
        action="store_true",
        help=f"Check if files require reformatting. "
        f"Return {SUCCESS} when there's nothing to reformat, "
        f"return {FAIL} when some files would be reformatted",
    )
    parser.add_argument(
        "-i",
        "--in-place",
        dest="in_place",
        default=False,
        action="store_true",
        help="Format files in-place",
    )
    parser.add_argument(
        "-l",
        "--line-length",
        metavar="INTEGER",
        dest="line_length",
        default=80,
        type=int,
        help="Maximum line length in characters",
    )
    parser.add_argument(
        "--definitions",
        dest="definitions",
        metavar="src",
        default=[],
        nargs="+",
        type=pathlib.Path,
        help="Files or directories containing custom command definitions (functions or macros). "
        "If only - is provided custom definitions, if there are any, are taken from stdin instead",
    )
    parser.add_argument(
        "--diff",
        dest="show_diff",
        default=False,
        action="store_true",
        help="Show diff on stdout for each formatted file instead",
    )
    parser.add_argument(
        "--unsafe",
        dest="format_safely",
        default=True,
        action="store_false",
        help="Skip default sanity checks",
    )
    parser.add_argument(
        dest="sources",
        metavar="src",
        nargs="*",
        type=pathlib.Path,
        help="File or directory to format. "
        "If only - is provided input is taken from stdin instead",
    )
    parser.add_argument(
        "--version",
        dest="show_version",
        default=False,
        action="store_true",
        help="Show version.",
    )
    return parser


def is_stdin_mixed_with_file_input(sources):
    return pathlib.Path("-") in sources and len(sources) != 1


def show_version():
    print(f"{__title__} {__version__}")
    print(f"lark {lark_version}")
    print(f"Python {sys.version}")


def main():
    argparser = create_argparser()
    args = argparser.parse_args()

    if args.show_version:
        show_version()
        sys.exit(SUCCESS)

    if any(map(is_stdin_mixed_with_file_input, [args.sources, args.definitions])):
        error("Don't mix stdin with file input")
        sys.exit(FAIL)

    sys.exit(run(args))


if __name__ == "__main__":
    main()



from contextlib import contextmanager
from dataclasses import dataclass
from difflib import unified_diff
from functools import lru_cache, partial
from itertools import chain
from pathlib import Path
import sys
import lark
from gersemi.exceptions import ASTMismatch, ParsingError
from gersemi.formatter import create_formatter
from gersemi.custom_command_dumper_generator import generate_custom_command_dumpers
from gersemi.parser import create_parser, create_parser_with_postprocessing


SUCCESS = 0
FAIL = 1
INTERNAL_ERROR = 123


error = partial(print, file=sys.stderr)


class StdinWrapper:  # pylint: disable=too-few-public-methods
    @staticmethod
    @lru_cache(maxsize=None)
    def read():
        return sys.stdin.read()


def standard_stream_open(mode):
    if mode is None or mode == "" or "r" in mode:
        return StdinWrapper()
    return sys.stdout


@contextmanager
def smart_open(filename, mode, *args, **kwargs):
    if filename == Path("-"):
        try:
            yield standard_stream_open(mode)
        finally:
            pass
    else:
        try:
            fh = open(filename, mode, *args, **kwargs)
            yield fh
        finally:
            fh.close()


def fromfile(path):
    return "<stdin>" if path == Path("-") else str(path)


def tofile(path):
    return "<stdout>" if path == Path("-") else str(path)


def get_newlines_style(code):
    crlf = "\r\n"
    if crlf in code:
        return crlf

    return "\n"


def translate_newlines_to_line_feed(code):
    return code.replace("\r\n", "\n").replace("\r", "\n")


def _print(txt, sink):
    print(txt, file=sink, end="")


def get_files(paths):
    def get_files_from_single_path(path):
        if path.is_dir():
            return chain(path.rglob("CMakeLists.txt"), path.rglob("*.cmake"),)
        return [path]

    for path in paths:
        for item in get_files_from_single_path(path):
            yield item


def has_custom_command_definition(code):
    lowercased = code.lower()
    has_function_definition = "function" in lowercased and "endfunction" in lowercased
    has_macro_definition = "macro" in lowercased and "endmacro" in lowercased
    return has_function_definition or has_macro_definition


def save_read(filepath):
    try:
        with smart_open(filepath, "r") as f:
            return f.read()
    except UnicodeDecodeError as exception:
        error(f"File {fromfile(filepath)} can't be read: ", exception)
        return None


def generate_specialized_dumpers(bare_parser, paths):
    parser = create_parser_with_postprocessing(bare_parser)
    result = dict()
    for filepath in get_files(paths):
        code = save_read(filepath)
        if code is None or not has_custom_command_definition(code):
            continue

        try:
            parse_tree = parser.parse(code)
            result.update(generate_custom_command_dumpers(parse_tree))
        except ParsingError as exception:
            error(f"{fromfile(filepath)}{exception}")
        except lark.exceptions.VisitError as exception:
            error(
                f"Runtime error when interpretting {fromfile(filepath)}: ", exception,
            )
    return result


@dataclass
class FormattedFile:
    before: str
    after: str
    newlines_style: str
    path: Path


def format_file(path, formatter):
    with smart_open(path, "r", newline="") as f:
        code = f.read()

    newlines_style = get_newlines_style(code)
    code = translate_newlines_to_line_feed(code)
    return FormattedFile(
        before=code,
        after=formatter.format(code),
        newlines_style=newlines_style,
        path=path,
    )


def show_diff(formatted_file):
    diff = unified_diff(
        a=f"{formatted_file.before}\n".splitlines(keepends=True),
        b=f"{formatted_file.after}\n".splitlines(keepends=True),
        fromfile=fromfile(formatted_file.path),
        tofile=tofile(formatted_file.path),
        n=5,
    )
    _print("".join(diff), sink=sys.stdout)
    return SUCCESS


def check_formatting(formatted_file):
    if formatted_file.before != formatted_file.after:
        error(f"{fromfile(formatted_file.path)} would be reformatted")
        return FAIL
    return SUCCESS


def format_in_place(formatted_file):
    with smart_open(
        formatted_file.path, "w", newline=formatted_file.newlines_style
    ) as f:
        _print(formatted_file.after, sink=f)
    return SUCCESS


def format_and_print_to_stdout(formatted_file):
    _print(formatted_file.after, sink=sys.stdout)
    return SUCCESS


def select_executor(args):
    if args.show_diff:
        return show_diff
    if args.check_formatting:
        return check_formatting
    if args.in_place:
        return format_in_place
    return format_and_print_to_stdout


def execute_on_single_file(file_to_format, formatter, execute):
    try:
        formatted_file = format_file(file_to_format, formatter)
        return execute(formatted_file)
    except ParsingError as exception:
        error(f"{fromfile(file_to_format)}{exception}")
    except ASTMismatch:
        error(
            f"Failed to format {fromfile(file_to_format)}: AST mismatch after formatting"
        )
    except lark.exceptions.VisitError as exception:
        error(f"Runtime error when formatting {fromfile(file_to_format)}: ", exception)
    return INTERNAL_ERROR


class LazyFormatter:  # pylint: disable=too-few-public-methods
    def __init__(self, args):
        self.args = args
        self.bare_parser = None
        self.formatter = None

    def _actual_format(self, code):
        return self.formatter.format(code)

    def format(self, code):
        if self.formatter is None:
            self.bare_parser = create_parser()
            self.formatter = create_formatter(
                self.bare_parser,
                self.args.format_safely,
                self.args.line_length,
                generate_specialized_dumpers(self.bare_parser, self.args.definitions),
            )

        return self._actual_format(code)


def run(args):
    formatter = LazyFormatter(args)
    run_on_single_file = partial(
        execute_on_single_file, formatter=formatter, execute=select_executor(args)
    )
    return max(map(run_on_single_file, get_files(args.sources)), default=SUCCESS)



class ASTMismatch(Exception):
    pass


class ParsingError(SyntaxError):
    description: str = ""

    def __str__(self):
        context, line, column = self.args  # pylint: disable=unbalanced-tuple-unpacking
        return f":{line}:{column}: {self.description}\n{context}"


class GenericParsingError(ParsingError):
    description = "unspecified parsing error"


class UnbalancedParentheses(ParsingError):
    description = "unbalanced parentheses"


class UnbalancedBrackets(ParsingError):
    description = "unbalanced brackets"



from itertools import dropwhile
import os
from typing import List
from lark import Discard, Tree
from lark.visitors import Transformer_InPlace
from gersemi.ast_helpers import is_newline
from gersemi.types import Nodes


class RemoveSuperfluousEmptyLines(Transformer_InPlace):
    def _drop_edge_empty_lines(self, children) -> List:
        while len(children) > 0 and is_newline(children[-1]):
            children.pop()
        return list(dropwhile(is_newline, children))

    def file(self, children) -> Tree:
        return Tree("file", self._drop_edge_empty_lines(children))

    def block_body(self, children) -> Tree:
        return Tree("block_body", self._drop_edge_empty_lines(children))


class SimplifyParseTree(Transformer_InPlace):
    def non_command_element(self, children: Nodes) -> Tree:
        if len(children) == 0:
            raise Discard
        return Tree("non_command_element", children)


def get_builtin_commands():
    HERE = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(HERE, "builtin_commands"), "r") as f:
        return set(f.read().splitlines())


class PreserveCustomCommandFormatting(Transformer_InPlace):
    builtin_commands = get_builtin_commands()

    def __init__(self, code):
        super().__init__()
        self.code = code

    def _get_original_formatting(self, lparen, rparen):
        start, end = lparen.end_pos, rparen.pos_in_stream
        content = self.code[start:end]
        return Tree("formatted_node", [content])

    def _get_indentation(self, identifier):
        end = identifier.pos_in_stream
        start = end - identifier.column + 1
        return self.code[start:end]

    def _make_custom_command(self, children):
        identifier, lparen, arguments, rparen = children
        indentation = self._get_indentation(identifier)
        return Tree(
            "custom_command",
            [
                indentation,
                identifier,
                arguments,
                self._get_original_formatting(lparen, rparen),
            ],
        )

    def _is_builtin(self, identifier):
        return identifier in self.builtin_commands

    def command_invocation(self, children):
        identifier, _, arguments, _ = children
        if self._is_builtin(identifier):
            return Tree("command_invocation", [identifier, arguments])

        return self._make_custom_command(children)


class PostProcessor(
    PreserveCustomCommandFormatting, SimplifyParseTree, RemoveSuperfluousEmptyLines
):
    pass



from contextlib import contextmanager
from lark import Tree
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.command_invocation_dumpers.ctest_command_dumpers import (
    ctest_command_mapping,
)
from gersemi.command_invocation_dumpers.scripting_command_dumpers import (
    scripting_command_mapping,
)
from gersemi.command_invocation_dumpers.project_command_dumpers import (
    project_command_mapping,
)
from gersemi.command_invocation_dumpers.preserving_command_invocation_dumper import (
    PreservingCommandInvocationDumper,
)


class CommandInvocationDumper(
    PreservingCommandInvocationDumper, BaseCommandInvocationDumper
):
    known_command_mapping = {
        **scripting_command_mapping,
        **project_command_mapping,
        **ctest_command_mapping,
    }

    @contextmanager
    def patched(self, patch):
        old_class = self.__class__

        class Impl(patch, old_class):  # pylint: disable=too-few-public-methods
            pass

        try:
            self.__class__ = Impl
            yield self
        finally:
            self.__class__ = old_class

    def _get_patch(self, command_name):
        return self.known_command_mapping.get(command_name, None)

    def command_invocation(self, tree):
        command_name, _ = tree.children
        patch = self._get_patch(command_name)
        if patch is None:
            return super().format_command(tree)
        with self.patched(patch):
            return self.format_command(tree)

    def custom_command(self, tree):
        _, command_name, arguments, *_ = tree.children
        if command_name in self.known_command_mapping:
            return self.visit(Tree("command_invocation", [command_name, arguments]))
        return super().custom_command(tree)



from typing import Dict, Iterator, Iterable, List, Optional, Sized, Tuple
from gersemi.ast_helpers import is_one_of_keywords
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.types import Nodes
from gersemi.utils import pop_all


def to_list_of_single_item_lists(sequence):
    return [*map(lambda item: [item], sequence)]


def is_non_empty(sequence: Sized) -> bool:
    return len(sequence) > 0


class ArgumentAwareCommandInvocationDumper(BaseCommandInvocationDumper):
    options: Iterable[str] = []
    one_value_keywords: Iterable[str] = []
    multi_value_keywords: Iterable[str] = []
    keyword_formatters: Dict[str, str] = {}

    def _default_format_values(self, values) -> str:
        return "\n".join(map(self.visit, values))

    def _format_group(self, group) -> str:
        result = self._try_to_format_into_single_line(group, separator=" ")
        if result is not None:
            return result

        keyword, *values = group
        begin = self.visit(keyword)
        if len(values) == 0:
            return begin

        keyword_as_value = keyword.children[0]
        with self.indented():
            formatter = getattr(
                self,
                self.keyword_formatters.get(keyword_as_value, "_default_format_values"),
            )
            formatted_values = formatter(values)
        return f"{begin}\n{formatted_values}"

    def _separate_front(self, arguments: Nodes) -> Tuple[List[Nodes], Nodes]:
        is_one_of_keywords_with_values = is_one_of_keywords(
            list(self.one_value_keywords) + list(self.multi_value_keywords)
        )
        for index, argument in enumerate(arguments):
            if is_one_of_keywords_with_values(argument):
                pivot = index
                break
        else:
            return to_list_of_single_item_lists(arguments), []
        return to_list_of_single_item_lists(arguments[:pivot]), arguments[pivot:]

    def _split_by_keywords(self, arguments: Nodes) -> Tuple[Iterator[Nodes], Nodes]:
        is_one_of_options = is_one_of_keywords(self.options)
        is_one_of_one_value_keywords = is_one_of_keywords(self.one_value_keywords)
        is_one_of_multi_value_keywords = is_one_of_keywords(self.multi_value_keywords)
        groups: List[Nodes] = []
        accumulator: Nodes = []
        tail: Optional[Nodes] = None
        iterator = iter(arguments)
        for argument in iterator:
            if is_one_of_options(argument):
                groups += [pop_all(accumulator)]
                groups += [[argument]]
            elif is_one_of_one_value_keywords(argument):
                groups += [pop_all(accumulator)]
                next_argument = next(iterator, None)
                if next_argument is None:
                    groups += [[argument]]
                else:
                    groups += [[argument, next_argument]]
            elif is_one_of_multi_value_keywords(argument):
                groups += [pop_all(accumulator)]
                accumulator = [argument]
            elif is_non_empty(accumulator):
                accumulator += [argument]
            else:
                tail = [argument, *iterator]
                break

        groups += [pop_all(accumulator)]
        if tail is None:
            tail = [*iterator]
        return filter(is_non_empty, groups), tail

    def _split_arguments(self, arguments: Nodes) -> List[Nodes]:
        front, tail = self._separate_front(arguments)
        keyworded_arguments, tail = self._split_by_keywords(tail)
        back = to_list_of_single_item_lists(tail)
        return [*front, *keyworded_arguments, *back]

    def format_command(self, tree):
        identifier, arguments = tree.children
        result = self._try_to_format_into_single_line(
            arguments.children, separator=" ", prefix=f"{identifier}(", postfix=")"
        )
        if result is not None:
            return result

        begin = self._indent(f"{identifier}(")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}\n{formatted_arguments}\n{end}"

    def arguments(self, tree):
        groups = self._split_arguments(tree.children)
        return "\n".join(map(self._format_group, groups))



from gersemi.configuration import Configuration
from gersemi.dumper import Dumper
from gersemi.parser import create_parser_with_postprocessing
from gersemi.sanity_checker import check_code_equivalence


def noop(*_):
    pass


class Formatter:  # pylint: disable=too-few-public-methods
    def __init__(
        self,
        bare_parser,
        sanity_checker,
        configuration: Configuration,
        custom_command_dumpers,
    ):
        self.bare_parser = bare_parser
        self.parser = create_parser_with_postprocessing(self.bare_parser)
        self.sanity_checker = sanity_checker
        self.configuration = configuration
        self.custom_command_dumpers = custom_command_dumpers

    def format(self, code):
        dumper = Dumper(self.configuration.line_length, self.custom_command_dumpers)
        result = dumper.visit(self.parser.parse(code))
        self.sanity_checker(self.bare_parser, code, result)
        return result


def create_formatter(bare_parser, do_sanity_check, line_length, custom_command_dumpers):
    sanity_checker = check_code_equivalence if do_sanity_check else noop
    return Formatter(
        bare_parser, sanity_checker, Configuration(line_length), custom_command_dumpers
    )



from gersemi.command_line_formatter import CommandLineFormatter
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)
from .command_with_property_value_pairs_dumper import (
    CommandWithPropertyValuePairsDumper,
)
from .condition_syntax_command_invocation_dumper import (
    ConditionSyntaxCommandInvocationDumper,
)
from .multiple_signature_command_invocation_dumper import (
    MultipleSignatureCommandInvocationDumper,
)
from .set_command import SetCommandDumper


class CMakeHostSysteInformationCommandDumper(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["RESULT"]
    multi_value_keywords = ["QUERY"]


class CMakeParseArgumentsCommandDumper(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["PARSE_ARGV"]


class ConfigureFileCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["COPYONLY", "ESCAPE_QUOTES", "@ONLY"]
    one_value_keywords = ["NEWLINE_STYLE"]


class EndForeachCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class EndFunctionCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class EndMacroCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class ExecuteProcessCommandDumper(
    CommandLineFormatter, ArgumentAwareCommandInvocationDumper
):
    options = [
        "OUTPUT_QUIET",
        "ERROR_QUIET",
        "OUTPUT_STRIP_TRAILING_WHITESPACE",
        "ERROR_STRIP_TRAILING_WHITESPACE",
    ]
    one_value_keywords = [
        "WORKING_DIRECTORY",
        "TIMEOUT",
        "RESULT_VARIABLE",
        "RESULTS_VARIABLE",
        "OUTPUT_VARIABLE",
        "ERROR_VARIABLE",
        "INPUT_FILE",
        "OUTPUT_FILE",
        "ERROR_FILE",
        "COMMAND_ECHO",
        "ENCODING",
    ]
    multi_value_keywords = ["COMMAND"]
    keyword_formatters = {"COMMAND": "_format_command_line"}


class FileCommandDumper(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        # Reading
        "READ": dict(
            options=["HEX"],
            one_value_keywords=["OFFSET", "LIMIT"],
            multi_value_keywords=["READ"],
        ),
        "STRINGS": dict(
            options=["NEWLINE_CONSUME", "NO_HEX_CONVERSION"],
            one_value_keywords=[
                "LENGTH_MAXIMUM",
                "LENGTH_MINIMUM",
                "LIMIT_COUNT",
                "LIMIT_INPUT",
                "LIMIT_OUTPUT",
                "REGEX",
                "ENCODING",
            ],
            multi_value_keywords=["STRINGS"],
        ),
        "GET_RUNTIME_DEPENDENCIES": dict(
            one_value_keywords=[
                "RESOLVED_DEPENDENCIES_VAR",
                "UNRESOLVED_DEPENDENCIES_VAR",
                "CONFLICTING_DEPENDENCIES_PREFIX",
                "BUNDLE_EXECUTABLE",
            ],
            multi_value_keywords=[
                "EXECUTABLES",
                "LIBRARIES",
                "MODULES",
                "DIRECTORIES",
                "PRE_INCLUDE_REGEXES",
                "PRE_EXCLUDE_REGEXES",
                "POST_INCLUDE_REGEXES",
                "POST_EXCLUDE_REGEXES",
            ],
        ),
        # Writing
        "GENERATE": dict(
            one_value_keywords=["INPUT", "CONTENT", "CONDITION"],
            multi_value_keywords=["GENERATE"],
        ),
        # Filesystem
        "GLOB": dict(
            options=["CONFIGURE_DEPENDS"],
            one_value_keywords=["GLOB", "LIST_DIRECTORIES", "RELATIVE"],
        ),
        "GLOB_RECURSE": dict(
            options=["CONFIGURE_DEPENDS"],
            one_value_keywords=["GLOB_RECURSE", "LIST_DIRECTORIES", "RELATIVE"],
        ),
        "COPY": dict(
            options=[
                "NO_SOURCE_PERMISSIONS",
                "USE_SOURCE_PERMISSIONS",
                "FOLLOW_SYMLINK_CHAIN",
                "FILES_MATCHING",
                "EXCLUDE",
            ],
            one_value_keywords=["DESTINATION", "PATTERN", "REGEX"],
            multi_value_keywords=[
                "COPY",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
                "PERMISSIONS",
            ],
        ),
        "INSTALL": dict(
            options=[
                "NO_SOURCE_PERMISSIONS",
                "USE_SOURCE_PERMISSIONS",
                "FOLLOW_SYMLINK_CHAIN",
                "FILES_MATCHING",
                "EXCLUDE",
            ],
            one_value_keywords=["DESTINATION", "PATTERN", "REGEX"],
            multi_value_keywords=[
                "INSTALL",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
                "PERMISSIONS",
            ],
        ),
        "CREATE_LINK": dict(
            options=["COPY_ON_ERROR", "SYMBOLIC"],
            one_value_keywords=["RESULT"],
            multi_value_keywords=["CREATE_LINK"],
        ),
        # Transfer
        "DOWNLOAD": dict(
            options=["SHOW_PROGRESS"],
            one_value_keywords=[
                "INACTIVITY_TIMEOUT",
                "LOG",
                "STATUS",
                "TIMEOUT",
                "USERPWD",
                "HTTPHEADER",
                "NETRC",
                "NETRC_FILE",
                "EXPECTED_HASH",
                "EXPECTED_MD5",
                "TLS_VERIFY",
                "TLS_CAINFO",
            ],
            multi_value_keywords=["DOWNLOAD"],
        ),
        "UPLOAD": dict(
            options=["SHOW_PROGRESS"],
            one_value_keywords=[
                "INACTIVITY_TIMEOUT",
                "LOG",
                "STATUS",
                "TIMEOUT",
                "USERPWD",
                "HTTPHEADER",
                "NETRC",
                "NETRC_FILE",
            ],
            multi_value_keywords=["UPLOAD"],
        ),
        # Locking
        "LOCK": dict(
            options=["DIRECTORY", "RELEASE"],
            one_value_keywords=["LOCK", "GUARD", "RESULT_VARIABLE", "TIMEOUT"],
        ),
    }


class FindFileCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
    ]
    one_value_keywords = ["DOC", "ENV"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindLibraryCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "NAMES_PER_DIR",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
    ]
    one_value_keywords = ["DOC", "ENV"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindPackageCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "EXACT",
        "QUIET",
        "MODULE",
        "CONFIG",
        "NO_MODULE",
        "NO_POLICY_SCOPE",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_PACKAGE_REGISTRY",
        "NO_CMAKE_BUILDS_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "NO_CMAKE_SYSTEM_PACKAGE_REGISTRY",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
    ]
    multi_value_keywords = [
        "REQUIRED",
        "COMPONENTS",
        "OPTIONAL_COMPONENTS",
        "NAMES",
        "CONFIGS",
        "HINTS",
        "PATHS",
        "PATH_SUFFIXES",
    ]


class FindPathCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
    ]
    one_value_keywords = ["DOC", "ENV"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindProgramCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "NAMES_PER_DIR",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
    ]
    one_value_keywords = ["DOC", "ENV"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class ForeachCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["IN"]
    multi_value_keywords = ["RANGE", "LISTS", "ITEMS"]


class FunctionCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class GetDirectoryPropertyCommandDumper(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DIRECTORY", "DEFINITION"]


class GetFilenameComponentCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "DIRECTORY",
        "NAME",
        "EXT",
        "NAME_WE",
        "LAST_EXT",
        "NAME_WLE",
        "PATH",
        "PROGRAM",
        "CACHE",
    ]
    one_value_keywords = ["BASE_DIR", "PROGRAM_ARGS"]


class GetPropertyCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["GLOBAL", "VARIABLE", "SET", "DEFINED", "BRIEF_DOCS", "FULL_DOCS"]
    one_value_keywords = [
        "TARGET",
        "SOURCE",
        "INSTALL",
        "TEST",
        "CACHE",
        "PROPERTY",
    ]
    multi_value_keywords = ["DIRECTORY"]


class IncludeCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["OPTIONAL", "NO_POLICY_SCOPE"]
    one_value_keywords = ["RESULT_VARIABLE"]


class ListCommandDumper(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        # Modification
        "FILTER": dict(
            options=["INCLUDE", "EXCLUDE"], one_value_keywords=["FILTER", "REGEX"],
        ),
        "REMOVE_DUPLICATES": dict(one_value_keywords=["REMOVE_DUPLICATES"],),
        "TRANSFORM": dict(
            one_value_keywords=["OUTPUT_VARIABLE"],
            multi_value_keywords=[
                "TRANSFORM",
                "APPEND",
                "PREPEND",
                "TOLOWER",
                "TOUPPER",
                "STRIP",
                "GENEX_STRIP",
                "REPLACE",
            ],
        ),
        # Ordering
        "SORT": dict(one_value_keywords=["SORT", "COMPARE", "CASE", "ORDER"],),
    }


class MacroCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class MarkAsAdvancedCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["CLEAR", "FORCE"]


class MathCommandDumper(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["OUTPUT_FORMAT"]
    multi_value_keywords = ["EXPR"]


class MessageCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = [
        "FATAL_ERROR",
        "SEND_ERROR",
        "WARNING",
        "AUTHOR_WARNING",
        "DEPRECATION",
        "NOTICE",
        "STATUS",
        "VERBOSE",
        "DEBUG",
        "TRACE",
    ]


class SeparateArgumentsCommandDumper(ArgumentAwareCommandInvocationDumper):
    pass


class SetPropertyCommandDumper(ArgumentAwareCommandInvocationDumper):
    options = ["GLOBAL", "APPEND", "APPEND_STRING"]
    one_value_keywords = ["DIRECTORY"]
    multi_value_keywords = ["TARGET", "SOURCE", "INSTALL", "TEST", "CACHE", "PROPERTY"]


class StringCommandDumper(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        # Search and Replace
        "FIND": dict(options=["REVERSE"], multi_value_keywords=["FIND"],),
        # Regular Expressions
        "REGEX": dict(one_value_keywords=["REGEX"],),
        # Comparison
        "COMPARE": dict(one_value_keywords=["COMPARE"],),
        # Generation
        "CONFIGURE": dict(
            options=["@ONLY", "ESCAPE_QUOTES"], multi_value_keywords=["CONFIGURE"],
        ),
        "RANDOM": dict(one_value_keywords=["LENGTH", "ALPHABET", "RANDOM_SEED"],),
        "UUID": dict(
            options=["UPPER"], one_value_keywords=["UUID", "NAMESPACE", "NAME", "TYPE"],
        ),
    }


scripting_command_mapping = {
    "cmake_host_system_information": CMakeHostSysteInformationCommandDumper,
    "cmake_parse_arguments": CMakeParseArgumentsCommandDumper,
    "configure_file": ConfigureFileCommandDumper,
    "elseif": ConditionSyntaxCommandInvocationDumper,
    "else": ConditionSyntaxCommandInvocationDumper,
    "endif": ConditionSyntaxCommandInvocationDumper,
    "endforeach": EndForeachCommandDumper,
    "endfunction": EndFunctionCommandDumper,
    "endmacro": EndMacroCommandDumper,
    "endwhile": ConditionSyntaxCommandInvocationDumper,
    "execute_process": ExecuteProcessCommandDumper,
    "file": FileCommandDumper,
    "find_file": FindFileCommandDumper,
    "find_library": FindLibraryCommandDumper,
    "find_package": FindPackageCommandDumper,
    "find_path": FindPathCommandDumper,
    "find_program": FindProgramCommandDumper,
    "foreach": ForeachCommandDumper,
    "function": FunctionCommandDumper,
    "get_directory_property": GetDirectoryPropertyCommandDumper,
    "get_filename_component": GetFilenameComponentCommandDumper,
    "get_property": GetPropertyCommandDumper,
    "if": ConditionSyntaxCommandInvocationDumper,
    "include": IncludeCommandDumper,
    "list": ListCommandDumper,
    "macro": MacroCommandDumper,
    "mark_as_advanced": MarkAsAdvancedCommandDumper,
    "math": MathCommandDumper,
    "message": MessageCommandDumper,
    "separate_arguments": SeparateArgumentsCommandDumper,
    "set_directory_properties": CommandWithPropertyValuePairsDumper,
    "set_property": SetPropertyCommandDumper,
    "set": SetCommandDumper,
    "string": StringCommandDumper,
    "while": ConditionSyntaxCommandInvocationDumper,
}



from lark import Token, Tree
from lark.visitors import Transformer
from gersemi.ast_helpers import is_newline
from gersemi.types import Nodes


class DowncaseIdentifiers(Transformer):
    def IDENTIFIER(self, token):
        return Token(
            "IDENTIFIER",
            token.lower(),
            pos_in_stream=token.pos_in_stream,
            line=token.line,
            column=token.column,
            end_line=token.end_line,
            end_column=token.end_column,
        )

    def else_term(self, token):
        return self.IDENTIFIER(token[0])

    elseif = else_term
    endforeach = else_term
    endfunction = else_term
    endif = else_term
    endmacro = else_term
    endwhile = else_term
    foreach = else_term
    function = else_term
    if_term = else_term
    macro = else_term
    while_term = else_term


class RestructureBracketTypeRules(Transformer):
    def _split_bracket_argument(self, arg):
        bracket_length = arg.index("[", 1) + 1
        return (
            arg[:bracket_length],
            arg[bracket_length:-bracket_length],
            arg[-bracket_length:],
        )

    def bracket_argument(self, children):
        token, *_ = children
        bracket_open, content, bracket_close = self._split_bracket_argument(token)
        return Tree(
            "bracket_argument",
            [
                Token("bracket_argument_begin", bracket_open),
                Token("bracket_argument_body", content),
                Token("bracket_argument_end", bracket_close),
            ],
        )

    def bracket_comment(self, children) -> Tree:
        *_, bracket_argument = children
        bracket_open, content, bracket_close = bracket_argument.children
        return Tree(
            "bracket_comment",
            [
                Token("bracket_comment_begin", f"#{bracket_open}"),
                Token("bracket_comment_body", content),
                Token("bracket_comment_end", bracket_close),
            ],
        )


class CleanUpComplexArgument(Transformer):
    def complex_argument(self, children):
        return Tree("complex_argument", children[1:-1])


class CleanUpNewlines(Transformer):
    def newline_or_gap(self, children):
        return Token("NEWLINE", "".join(children)[:2])

    def arguments(self, children: Nodes) -> Tree:
        return Tree("arguments", [child for child in children if not is_newline(child)])


class ParsingTransformer(
    DowncaseIdentifiers,
    RestructureBracketTypeRules,
    CleanUpComplexArgument,
    CleanUpNewlines,
):
    pass



import re
from textwrap import indent
from gersemi.base_dumper import BaseDumper


BRACKET_ARGUMENT_REGEX = r"(\[(?P<equal_signs>(=*))\[(?:[\s\S]+?)\](?P=equal_signs)\])"
QUOTED_ARGUMENT_REGEX = r'("(?:[^\\\"]|\n|(?:\\(?:[^A-Za-z0-9]|[nrt]))|\\\n)*")'


def split_by_bracket_arguments(string):
    return re.split(BRACKET_ARGUMENT_REGEX, string)


def split_by_quoted_arguments(string):
    return re.split(QUOTED_ARGUMENT_REGEX, string)


def split_into_segments(string):
    segments = split_by_bracket_arguments(string)
    result = [split_by_quoted_arguments(segment) for segment in segments]
    return [item for segment in result for item in segment if item != ""]


def indent_segment(segment, indent_size):
    prefix = " " * indent_size
    if segment[:1] in ["[", '"']:
        return f"{prefix}{segment}"
    return indent(segment, prefix)


def safe_indent(string, indent_size):
    segments = split_into_segments(string)
    segment_indenter = lambda segment: indent_segment(segment, indent_size)
    return "".join(map(segment_indenter, segments))


class PreservingCommandInvocationDumper(BaseDumper):
    def custom_command(self, tree):
        indentation, identifier, _, formatted_arguments = tree.children
        begin = self._indent(f"{identifier}(")
        content = formatted_arguments.children[0]
        if content == "":
            return f"{begin})"

        with self.not_indented():
            result = self._try_to_format_into_single_line(
                formatted_arguments.children, separator=" ", prefix=begin, postfix=")"
            )
        if result is not None:
            return result

        body = safe_indent(content, self.alignment - len(indentation))
        if not content.startswith("\n"):
            body = body.strip(" ")

        end = self._indent(")") if "\n" in body else ")"
        return f"{begin}{body}{end}"



from itertools import filterfalse
from typing import Dict, List, Union
from gersemi.ast_helpers import is_comment, is_unquoted_argument
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class MultipleSignatureCommandInvocationDumper(ArgumentAwareCommandInvocationDumper):
    customized_signatures: Dict[str, Dict[str, Union[List, int]]] = {}

    def _update_signature_characteristics(self, signature):
        self.options = signature.get("options", [])
        self.one_value_keywords = signature.get("one_value_keywords", [])
        self.multi_value_keywords = signature.get("multi_value_keywords", [])

    def arguments(self, tree):
        arguments_only = filterfalse(is_comment, tree.children)
        first_argument = next(arguments_only, None)
        if first_argument is None or not is_unquoted_argument(first_argument):
            return super().arguments(tree)
        first_argument_as_value = first_argument.children[0]
        signature = self.customized_signatures.get(first_argument_as_value, None)
        if signature is not None:
            self._update_signature_characteristics(signature)
        return super().arguments(tree)



from typing import Callable
from lark import Tree, Token
from lark.visitors import Interpreter
from gersemi.types import Node


def is_tree(tree_type: str) -> Callable[[Node], bool]:
    return lambda element: isinstance(element, Tree) and element.data == tree_type


def is_token(token_type: str) -> Callable[[Node], bool]:
    return lambda element: isinstance(element, Token) and element.type == token_type


is_line_comment = is_tree("line_comment")
is_bracket_comment = is_tree("bracket_comment")
is_unquoted_argument = is_tree("unquoted_argument")
is_commented_argument = is_tree("commented_argument")
is_block = is_tree("block")

is_newline = is_token("NEWLINE")

is_comment = lambda element: is_bracket_comment(element) or is_line_comment(element)


class ContainsLineComment(Interpreter):
    def __default__(self, _) -> bool:
        return False

    def line_comment(self, _) -> bool:
        return True

    def _visit(self, tree: Tree) -> bool:
        is_subtree = lambda node: isinstance(node, Tree)
        subtrees = filter(is_subtree, tree.children)
        return any(map(self.visit, subtrees))

    arguments = _visit
    commented_argument = _visit


def contains_line_comment(nodes) -> bool:
    visit = ContainsLineComment().visit
    check_node = lambda node: isinstance(node, Tree) and visit(node)
    return any(map(check_node, nodes))


def is_commented_unquoted_argument(node):
    return is_commented_argument(node) and is_unquoted_argument(node.children[0])


def is_keyword(keyword):
    def impl(node):
        return (is_unquoted_argument(node) and node.children[0] == keyword) or (
            is_commented_unquoted_argument(node)
            and node.children[0].children[0] == keyword
        )

    return impl


class KeywordMatcher:  # pylint: disable=too-few-public-methods
    def __init__(self, keywords):
        self.keywords = keywords

    def __eq__(self, other):
        return any(map(other.__eq__, self.keywords))


def is_one_of_keywords(keywords):
    return is_keyword(KeywordMatcher(keywords))



from contextlib import contextmanager
from textwrap import indent
from typing import Optional
from lark import Tree
from lark.visitors import Interpreter
from gersemi.ast_helpers import contains_line_comment
from gersemi.types import Nodes


class BaseDumper(Interpreter):
    def __init__(self, width, alignment=0):
        self.width = width
        self.indent_size = 4
        self.alignment = alignment

    def __default__(self, tree: Tree):
        return "".join(self.visit_children(tree))

    def _indent(self, text: str):
        return indent(text, " " * self.alignment)

    def _try_to_format_into_single_line(
        self, children: Nodes, separator: str = "", prefix: str = "", postfix: str = ""
    ) -> Optional[str]:
        if not contains_line_comment(children):
            with self.not_indented():
                formatted_children = separator.join(
                    self.visit(c) if isinstance(c, Tree) else c for c in children
                )
            result = self._indent(f"{prefix}{formatted_children}{postfix}")
            if len(result) <= self.width and "\n" not in result:
                return result
        return None

    @contextmanager
    def aligned_to(self, alignment):
        try:
            old_alignment = self.alignment
            self.alignment = alignment
            yield self
        finally:
            self.alignment = old_alignment

    def indented(self):
        return self.aligned_to(self.alignment + self.indent_size)

    def not_indented(self):
        return self.aligned_to(0)



from dataclasses import dataclass
from typing import Tuple
from lark import Discard, Tree
from lark.visitors import Interpreter, Transformer
from gersemi.ast_helpers import is_keyword
from gersemi.command_invocation_dumpers.argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class DropIrrelevantElements(Transformer):  # pylint: disable=too-few-public-methods
    def _discard(self, _):
        raise Discard

    NEWLINE = _discard
    line_comment = _discard
    bracket_comment = _discard
    non_command_element = _discard


@dataclass
class Keywords:
    options: Tuple[str]
    one_value_keywords: Tuple[str]
    multi_value_keywords: Tuple[str]


def create_specialized_dumper(keywords: Keywords):
    class Impl(ArgumentAwareCommandInvocationDumper):
        options = keywords.options
        one_value_keywords = keywords.one_value_keywords
        multi_value_keywords = keywords.multi_value_keywords

        def custom_command(self, tree):
            _, command_name, arguments, *_ = tree.children
            return self.visit(Tree("command_invocation", [command_name, arguments]))

    return Impl


is_parse_argv = is_keyword("PARSE_ARGV")


class CMakeInterpreter(Interpreter):
    def __init__(self, stack=None):
        self.stack = dict() if stack is None else stack
        self.found_commands = dict()

    @property
    def _inner_scope(self):
        return type(self)(stack=self.stack.copy())

    def _eval_variables(self, arg):
        for name, value in self.stack.items():
            arg = arg.replace(f"${{{name}}}", ";".join(value))
        return arg.split(";")

    def _set(self, arguments):
        name, *values = self.visit_children(arguments)
        self.stack[name] = [
            item for value in values for item in self._eval_variables(value)
        ]

    def _new_command(self, arguments):
        if len(arguments.children) > 0:
            name = arguments.children[0]
            return self.visit(name)
        raise RuntimeError

    def _cmake_parse_arguments(self, arguments):
        if is_parse_argv(arguments.children[0]):
            keywords = arguments.children[3:6]
        else:
            keywords = arguments.children[1:4]

        options, one_value_keywords, multi_value_keywords = [
            self._eval_variables(self.visit(item)) for item in keywords
        ]
        return Keywords(options, one_value_keywords, multi_value_keywords)

    def file(self, tree):
        self.visit_children(tree)
        return self.found_commands

    def block(self, tree):
        subinterpreter = self._inner_scope
        block_begin, *body, _ = subinterpreter.visit_children(tree)
        name, *_ = block_begin
        if name is None:
            return

        keywords, *_ = body
        if len(keywords) > 0:
            self.found_commands[name.lower()] = keywords[0]

    def block_body(self, tree):
        return [
            item for item in self.visit_children(tree) if isinstance(item, Keywords)
        ][:1]

    def command_invocation(self, tree):
        identifier, arguments = tree.children
        command_interpreters = {
            "cmake_parse_arguments": self._cmake_parse_arguments,
            "function": self._new_command,
            "macro": self._new_command,
            "set": self._set,
        }
        return command_interpreters.get(identifier, lambda *args: None)(arguments)

    def _join(self, tree):
        return "".join(self.visit_children(tree))

    def complex_argument(self, tree):
        return f"({self.visit_children(tree)})"

    def _extract_first(self, tree):
        return tree.children[0]

    def quoted_argument(self, tree):
        return tree.children[0] if tree.children else ""

    bracket_argument = _join
    commented_argument = _join
    unquoted_argument = _extract_first


class GenerateCustomCommandDumpers(Interpreter):
    def start(self, tree):
        return self.visit_children(tree)[0]

    def file(self, tree):
        tree = DropIrrelevantElements(visit_tokens=True).transform(tree)
        found_commands = CMakeInterpreter().visit(tree)
        return {
            name: create_specialized_dumper(keywords)
            for name, keywords in found_commands.items()
        }

    def __default__(self, tree):
        return None


def generate_custom_command_dumpers(tree):
    return GenerateCustomCommandDumpers().visit(tree)



__author__ = "Blank Spruce"
__author_email__ = "blankspruce@protonmail.com"
__description__ = "A formatter to make your CMake code the real treasure"
__license__ = "MPL 2.0"
__title__ = "gersemi"
__url__ = "https://github.com/BlankSpruce/gersemi"
__version__ = "0.2.1"
